//! \file gnulliver.c
//! Gnullify the big-endian / little-endian controversy.
//==============================================================================
// AEPHEM - an astronomical ephemeris and reduction library.
// Copyright 2008 Adam Hincks.
//
// This file is part of AEPHEM.
//
// AEPHEM is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AEPHEM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AEPHEM.  If not, see <http://www.gnu.org/licenses/>.
//==============================================================================
// This file is based on code from the ephcom library, which is distributed 
// under the Lesser Gnu Public License.  As per paragraph 7 of version 2.1 of
// that licence, the intact library is included in the AEPHEM library
// distribution in the directory src/ephcom.
//==============================================================================

#include "gnulliver.h"

//------------------------------------------------------------------------------
//! Determine endianness.
//!
//! \return One of #AE_GNULLIVER_BIG, #AE_GNULLIVER_LITTLE or #AE_GNULLIVER_DEC.
//------------------------------------------------------------------------------

unsigned char ae_gnulliver() {
  unsigned char result;
  union {
    double d;
    unsigned char ch[8];
  } bytes;

  bytes.d = AE_GNULLIVER_TEST;
  result = 0;

  if (bytes.ch[0] == 0) {      // Swop 32-bit words in a 64-bit number.
    if (bytes.ch[4] == 0x10)               // Assume 0x00000000 10203040.
      result = AE_GNULLIVER_LITTLE;
    else                                   // Assume 0x00000000 20104030.
      result = AE_GNULLIVER_DEC;
  }
  else                                      // Assume 0x40302010 00000000.
    result = AE_GNULLIVER_BIG;

  return(result);
}


//------------------------------------------------------------------------------
//! Do any necessary swopping for a 16-bit integer.
//! Endianness of the machine is automatically detected.
//!
//! \param input The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

unsigned short ae_gnulliver16(unsigned short input) {
  static unsigned endian, init = 1;
  unsigned char tmpch;
  union {
    unsigned short u;
    unsigned char ch[2];
  } bytes;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  if (endian != AE_GNULLIVER_BIG) {
    tmpch = bytes.ch[0]; 
    bytes.ch[0] = bytes.ch[1]; 
    bytes.ch[1] = tmpch;
  }

  return bytes.u;
}


//------------------------------------------------------------------------------
//! Do any necessary swopping for a 32-bit integer.
//! Endianness of the machine is automatically detected.
//!
//! \param input The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

unsigned ae_gnulliver32(unsigned input) {
  static unsigned endian, init = 1;
  char tmpch;
  union {
    unsigned u;
    unsigned char ch[4];
  } bytes;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  bytes.u = input;
  if (endian != AE_GNULLIVER_BIG) {
    if (endian == AE_GNULLIVER_LITTLE) {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[3]; 
      bytes.ch[3] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[2]; 
      bytes.ch[2] = tmpch;
    }
    else {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[2]; 
      bytes.ch[2] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[3]; 
      bytes.ch[3] = tmpch;
    }
  }

  return bytes.u;
}


//------------------------------------------------------------------------------
//! Do any necessary swopping for a 64-bit integer.
//! Endianness of the machine is automatically detected.
//!
//! \param input The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

unsigned long ae_gnulliver64(unsigned long input) {
  static unsigned endian, init = 1;
  char tmpch;
  union {
    unsigned long u;
    char ch[8];
  } bytes;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  bytes.u = input;
  if (endian != AE_GNULLIVER_BIG) {
    if (endian == AE_GNULLIVER_LITTLE) {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[7]; 
      bytes.ch[7] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[6]; 
      bytes.ch[6] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[5]; 
      bytes.ch[5] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[4]; 
      bytes.ch[4] = tmpch;
    }
    else {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[6]; 
      bytes.ch[6] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[7]; 
      bytes.ch[7] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[4]; 
      bytes.ch[4] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[5]; 
      bytes.ch[5] = tmpch;
    }
  }

  return bytes.u;
}


//------------------------------------------------------------------------------
//! Do any necessary swopping for a 64-bit character string.
//! Endianness of the machine is automatically detected.
//!
//! \param ch The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

unsigned char *ae_gnulliver64c(unsigned char *ch) {
  static unsigned endian, init = 1;
  unsigned char tmpch;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  if (endian != AE_GNULLIVER_BIG) {
    if (endian == AE_GNULLIVER_LITTLE) {
      tmpch = ch[0]; 
      ch[0] = ch[7]; 
      ch[7] = tmpch;
      tmpch = ch[1]; 
      ch[1] = ch[6]; 
      ch[6] = tmpch;
      tmpch = ch[2]; 
      ch[2] = ch[5]; 
      ch[5] = tmpch;
      tmpch = ch[3]; 
      ch[3] = ch[4]; 
      ch[4] = tmpch;
    }
    else {
      tmpch = ch[0]; 
      ch[0] = ch[6]; 
      ch[6] = tmpch;
      tmpch = ch[1]; 
      ch[1] = ch[7]; 
      ch[7] = tmpch;
      tmpch = ch[2]; 
      ch[2] = ch[4]; 
      ch[4] = tmpch;
      tmpch = ch[3]; 
      ch[3] = ch[5]; 
      ch[5] = tmpch;
    }
  }

   return ch;
}


//------------------------------------------------------------------------------
//! Do any necessary swopping for a 128-bit integer.
//! Endianness of the machine is automatically detected.
//!
//! \param input The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

unsigned long long ae_gnulliver128(unsigned long long input) {
  static unsigned endian, init = 1;
  unsigned char tmpch;
  union {
    unsigned long long u;
    char ch[16];
  } bytes;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  bytes.u = input;
  if (endian != AE_GNULLIVER_BIG) {
    if (endian == AE_GNULLIVER_LITTLE) {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[15]; 
      bytes.ch[15] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[14]; 
      bytes.ch[14] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[13]; 
      bytes.ch[13] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[12]; 
      bytes.ch[12] = tmpch;
      tmpch = bytes.ch[4]; 
      bytes.ch[4] = bytes.ch[11]; 
      bytes.ch[11] = tmpch;
      tmpch = bytes.ch[5]; 
      bytes.ch[5] = bytes.ch[10]; 
      bytes.ch[10] = tmpch;
      tmpch = bytes.ch[6]; 
      bytes.ch[6] = bytes.ch[ 9]; 
      bytes.ch[ 9] = tmpch;
      tmpch = bytes.ch[7]; 
      bytes.ch[7] = bytes.ch[ 8]; 
      bytes.ch[ 8] = tmpch;
    }
    else {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[14]; 
      bytes.ch[14] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[15]; 
      bytes.ch[15] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[12]; 
      bytes.ch[12] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[13]; 
      bytes.ch[13] = tmpch;
      tmpch = bytes.ch[4]; 
      bytes.ch[4] = bytes.ch[10]; 
      bytes.ch[10] = tmpch;
      tmpch = bytes.ch[5]; 
      bytes.ch[5] = bytes.ch[11]; 
      bytes.ch[11] = tmpch;
      tmpch = bytes.ch[6]; 
      bytes.ch[6] = bytes.ch[8]; 
      bytes.ch[8] = tmpch;
      tmpch = bytes.ch[7]; 
      bytes.ch[7] = bytes.ch[9];
      bytes.ch[9] = tmpch;
    }
  }

  return bytes.u;
}


//------------------------------------------------------------------------------
//! Do any necessary swopping for a 32-bit float.
//! Endianness of the machine is automatically detected.
//!
//! \param input32 The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

float ae_gnulliver32f(float input32) {
  static unsigned endian, init = 1;
  unsigned char tmpch;
  union {
    float x32;
    unsigned char ch[4];
  } bytes;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  bytes.x32 = input32;
  if (endian != AE_GNULLIVER_BIG) {
    if (endian == AE_GNULLIVER_LITTLE) {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[3]; 
      bytes.ch[3] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[2]; 
      bytes.ch[2] = tmpch;
    }
    else {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[2]; 
      bytes.ch[2] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[3]; 
      bytes.ch[3] = tmpch;
    }
  }

  return bytes.x32;
}


//------------------------------------------------------------------------------
//! Do any necessary swopping for a 64-bit float
//! Endianness of the machine is automatically detected.
//!
//! \param input64 The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

double ae_gnulliver64f(double input64) {
  static unsigned endian, init=1;
  unsigned char tmpch;
  union {
    double x64;
    unsigned char ch[8];
  } bytes;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  bytes.x64 = input64;
  if (endian != AE_GNULLIVER_BIG) {
    if (endian == AE_GNULLIVER_LITTLE) {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[7]; 
      bytes.ch[7] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[6]; 
      bytes.ch[6] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[5]; 
      bytes.ch[5] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[4]; 
      bytes.ch[4] = tmpch;
    }
    else {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[6]; 
      bytes.ch[6] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[7]; 
      bytes.ch[7] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[4]; 
      bytes.ch[4] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[5]; 
      bytes.ch[5] = tmpch;
    }
  }

  return bytes.x64;
}

//------------------------------------------------------------------------------
//! Do any necessary swopping for a 16-bit number.
//! Endianness of the machine is automatically detected.  Note: This function's 
//! result will be unpredictable if the type long double is not a 128-bit word.
//!
//! \param input128 The raw input.
//!
//! \return The number in correct byte-order.
//------------------------------------------------------------------------------

long double ae_gnulliver128f(long double input128) {
  static unsigned endian, init=1;
  unsigned char tmpch;
  union {
    long double x128;
    unsigned char ch[16];
  } bytes;

  if (init) {
    endian = ae_gnulliver();
    init = 0;
  }
  bytes.x128 = input128;
  if (endian != AE_GNULLIVER_BIG) {
    if (endian == AE_GNULLIVER_LITTLE) {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[15]; 
      bytes.ch[15] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[14]; 
      bytes.ch[14] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[13]; 
      bytes.ch[13] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[12]; 
      bytes.ch[12] = tmpch;
      tmpch = bytes.ch[4]; 
      bytes.ch[4] = bytes.ch[11]; 
      bytes.ch[11] = tmpch;
      tmpch = bytes.ch[5]; 
      bytes.ch[5] = bytes.ch[10]; 
      bytes.ch[10] = tmpch;
      tmpch = bytes.ch[6]; 
      bytes.ch[6] = bytes.ch[9]; 
      bytes.ch[9] = tmpch;
      tmpch = bytes.ch[7]; 
      bytes.ch[7] = bytes.ch[8]; 
      bytes.ch[8] = tmpch;
    }
    else {
      tmpch = bytes.ch[0]; 
      bytes.ch[0] = bytes.ch[14]; 
      bytes.ch[14] = tmpch;
      tmpch = bytes.ch[1]; 
      bytes.ch[1] = bytes.ch[15]; 
      bytes.ch[15] = tmpch;
      tmpch = bytes.ch[2]; 
      bytes.ch[2] = bytes.ch[12]; 
      bytes.ch[12] = tmpch;
      tmpch = bytes.ch[3]; 
      bytes.ch[3] = bytes.ch[13]; 
      bytes.ch[13] = tmpch;
      tmpch = bytes.ch[4]; 
      bytes.ch[4] = bytes.ch[10]; 
      bytes.ch[10] = tmpch;
      tmpch = bytes.ch[5]; 
      bytes.ch[5] = bytes.ch[11]; 
      bytes.ch[11] = tmpch;
      tmpch = bytes.ch[6]; 
      bytes.ch[6] = bytes.ch[8]; 
      bytes.ch[ 8] = tmpch;
      tmpch = bytes.ch[7]; 
      bytes.ch[7] = bytes.ch[9]; 
      bytes.ch[ 9] = tmpch;
    }
  }

  return bytes.x128;
}
