//! \file gnulliver.h
//! Gnullify the big-endian / little-endian controversy.
//! Swops values between host order and network order without needing any 
//! include files and without calls to htonl(), etc.  This is a symmetric 
//! conversion, so the same function works in both directions.  For example, to 
//! convert an integer between host byte ordering and network byte ordering 
//! (in either direction), use
//! \code 
//!   int i, j;
//!   j = gnulliver32(i);
//! \endcode
//! then use j in an equation (if you just read i from the net), or
//! write j to the net (if i is an integer on your machine).
//!
//! The gnulliver routines work by setting a double to a certain value, then 
//! examining a unioned byte array.
//!   
//! The code in the routines might appear largely redundant, and it is.  Loops 
//! are also unrolled.  All this should minimize function calls and maximize 
//! speed.
//==============================================================================
// AEPHEM - an astronomical ephemeris and reduction library.
// Copyright 2008 Adam Hincks.
//
// This file is part of AEPHEM.
//
// AEPHEM is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AEPHEM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AEPHEM.  If not, see <http://www.gnu.org/licenses/>.
//==============================================================================
// This file is based on code from the ephcom library, which is distributed 
// under the Lesser Gnu Public License.  As per paragraph 7 of version 2.1 of
// that licence, the intact library is included in the AEPHEM library
// distribution in the directory src/ephcom.
//==============================================================================

#ifndef GNULLIVER_H
#define GNULLIVER_H

#ifdef __cplusplus
  extern "C" {
#endif

#define AE_GNULLIVER_SWAP16  1   //!< Swop adjacent bytes in 2-byte string.
#define AE_GNULLIVER_SWAP32  2   //!< Swop 2-byte pairs in 4-byte string.
#define AE_GNULLIVER_SWAP64  4   //!< Swop 4-byte pairs in 8-byte string.
#define AE_GNULLIVER_SWAP128 8   //!< Swop 8-byte pairs in 16-byte string.

#define AE_GNULLIVER_BIG     0   //!< No swopping if this machine is Big-endian.
//! Swopping required.
#define AE_GNULLIVER_LITTLE (AE_GNULLIVER_SWAP16 | AE_GNULLIVER_SWAP32 | \
                             AE_GNULLIVER_SWAP64 | AE_GNULLIVER_SWAP128)
//! This should work on a PDP-11 or VAX, but has not been tested.
#define AE_GNULLIVER_DEC    (AE_GNULLIVER_SWAP32 | AE_GNULLIVER_SWAP64 | \
                             AE_GNULLIVER_SWAP128)

//! A magic number for determining the byte swapping in a 64-bit word.
//! A IEEE 754 floating point is assumed, which to hexadecimal 0x40302010 
//! 00000000 on a Big-endian (network order) machine. The mantissa is large 
//! enough that it might work on a non-IEEE 754 compatible machine just by 
//! coincidence, but no guarantees.
//!
//! Gnulliver currently assumes that if we swap 32-bit values in a 64-bit word,
//! then we also swap 64-bit values in a 128-bit word (long double), which is
//! true if the machine is Little-endian.
#define AE_GNULLIVER_TEST (16.0 + (1.0/8.0 + 1.0/4096.0))


unsigned char ae_gnulliver();
unsigned short ae_gnulliver16(unsigned short input);
unsigned ae_gnulliver32(unsigned input);
unsigned long ae_gnulliver64(unsigned long input);
unsigned char *ae_gnulliver64c(unsigned char *ch);
unsigned long long ae_gnulliver128(unsigned long long input);
float ae_gnulliver32f(float input);
double ae_gnulliver64f(double input);
long double ae_gnulliver128f(long double input);

#ifdef __cplusplus
  }
#endif

#endif
