//! \file jpl_eph.c
//! Read JPL ephemeris files.
//==============================================================================
// AEPHEM - an astronomical ephemeris and reduction library.
// Copyright 2008 Adam Hincks.
//
// This file is part of AEPHEM.
//
// AEPHEM is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AEPHEM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AEPHEM.  If not, see <http://www.gnu.org/licenses/>.
//==============================================================================
// This file is based on code from the ephcom library, which is distributed 
// under the Lesser Gnu Public License.  As per paragraph 7 of version 2.1 of
// that licence, the intact library is included in the AEPHEM library
// distribution in the directory src/ephcom.
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "aephem.h"
#include "gnulliver.h"

#define AE_JPL_MAXLINE     128  //!< The maximum number of characters for input.
#define AE_JPL_TTL_MAXLINE  86  //!< The maximum length of TTL input.

int ae_jpl_ascii_next_group(char *group, char *expected, FILE *f);
int ae_jpl_read_ascii_header(const char *path, struct ae_jpl_handle_t *jh);
int ae_jpl_read_bin_header(const char *path, struct ae_jpl_handle_t *jh);
int ae_jpl_read_ascii_block(struct ae_jpl_handle_t *jh);
int ae_jpl_read_bin_block(struct ae_jpl_handle_t *jh, int blocknum);
int ae_jpl_cheby(struct ae_jpl_handle_t *jh, double jd_tt, int obj_num, 
                 double *pv, char is_planetary);
int ae_jpl_get_double(FILE *fp, double *val);
int ae_jpl_get_int(FILE *fp, int *val);


//------------------------------------------------------------------------------
//! Initialise to use a JPL ephemeris file.
//! This function opens the file, determines whether it is binary or ASCII, and
//! parses the header information.  This information is stored in the handle \p
//! jh and is passed to subsequent JPL function calls.
//!
//! <em>Note that ASCII files have not been fully tested and might not
//! work.</em>
//!
//! \param path The location of the ephemeris file.
//! \param jh For returning a handle.
//!
//! \return 0 on success; otherwise, a negative code from #ae_retcode_t.
//------------------------------------------------------------------------------

int ae_jpl_init(const char *path, struct ae_jpl_handle_t *jh) {
  int i;

  jh->fp = NULL;

  if (ae_jpl_read_ascii_header(path, jh)) {
    if ((i = ae_jpl_read_bin_header(path, jh))) {
      jh->fp = NULL;
      return i;
    }
    jh->is_bin = 1;
  }
  else
    jh->is_bin = 0;

  jh->pc = (double *)malloc(jh->maxcheby * sizeof(double));
  jh->vc = (double *)malloc(jh->maxcheby * sizeof(double));
  jh->datablock = (double *)malloc(jh->ncoeff * jh->maxcheby * sizeof(double));

  return 0;
}


//------------------------------------------------------------------------------
//! Finish with a JPL handle.
//! Closes any open file pointers and frees any allocated arrays.
//!
//! \param jh A JPL handle.
//------------------------------------------------------------------------------

void ae_jpl_close(struct ae_jpl_handle_t *jh) {
  if (jh->fp != NULL) {
    fclose(jh->fp);
    free(jh->pc);
    free(jh->vc);
    free(jh->datablock);
  }

  return;
}


//------------------------------------------------------------------------------
//! Read a JPL ephemeris ASCII header.
//!
//! \param path The location of the ephemeris file.
//! \param jh For returning the header information.
//!
//! \return 0 on success; on error, a negative error code from #ae_retcode_t.
//------------------------------------------------------------------------------

int ae_jpl_read_ascii_header(const char *path, struct ae_jpl_handle_t *jh) {
  char group[13], readbuf[AE_JPL_MAXLINE + 1];;
  int i, j, iword, blockbytes;

  if ((jh->fp = fopen(path, "r")) == NULL)
    return -AE_RET_BAD_PATH;

  // First header line: KSIZE= # NCOEFF= #.
  if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp) == NULL) {
    fclose(jh->fp);
    return -AE_RET_UNEXPECTED_EOF;
  }
  sscanf(readbuf, "%*6s%6d%*11s%6d", &jh->ksize, &jh->ncoeff);
  blockbytes = 8 * jh->ncoeff;  // Double times # of doubles/block.
  if (jh->ksize != 2 * jh->ncoeff) {
    fclose(jh->fp);
    return -AE_RET_BAD_JPL_HEADER_KSIZE;
  }

  // GROUP 1010: Title of ephemeris (DE/LE number, start JD, end JD).
  if ((i = ae_jpl_ascii_next_group(group, "GROUP   1010", jh->fp))) {
    fclose(jh->fp);
    return i;
  }

  // Title line.
  if (fgets(jh->ttl[0], AE_JPL_TTL_MAXLINE, jh->fp) == NULL) {
    fclose(jh->fp);
    return -AE_RET_UNEXPECTED_EOF;
  }
  if (strncmp(jh->ttl[0], "JPL ", 4)) {
    fclose(jh->fp);
    return -AE_RET_BAD_JPL_NOT_ASCII;
  }

  // Start and end epoch.
  if (fgets(jh->ttl[1], AE_JPL_TTL_MAXLINE, jh->fp) == NULL) {
    fclose(jh->fp);
    return -AE_RET_UNEXPECTED_EOF;
  }
  if (fgets(jh->ttl[2], AE_JPL_TTL_MAXLINE, jh->fp) == NULL) {
    fclose(jh->fp);
    return -AE_RET_UNEXPECTED_EOF;
  }

  // Convert any newlines or tabs to single spaces.
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 84; j++) {
      if (isspace(jh->ttl[i][j]))
        jh->ttl[i][j] = ' ';
    }
    jh->ttl[i][84] = '\0';
  }

  // GROUP 1030: Start and End JD, timestep (in JD) per block.
  if ((i = ae_jpl_ascii_next_group(group, "GROUP   1030", jh->fp))) {
    fclose(jh->fp);
    return i;
  }
  if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp) == NULL) {
    fclose(jh->fp);
    return -AE_RET_UNEXPECTED_EOF;
  }
  sscanf(readbuf, " %lE %lE %lE", &jh->ss[0], &jh->ss[1], 
         &jh->ss[2]);

  // GROUP 1040: Constant names.
  if ((i = ae_jpl_ascii_next_group(group, "GROUP   1040", jh->fp))) {
    fclose(jh->fp);
    return i;
  }
  if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp)) {
    fclose(jh->fp);
    return i;
  }
  jh->ncon = atoi(readbuf);

  // Now read the constant names, 10 per line, each 6 characters long preceded
  // by 2 blanks.  Pad names with blanks to make 6 characters.
  for (i = 0; i < jh->ncon;) {
    if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp) == NULL) {
      fclose(jh->fp);
      return -AE_RET_UNEXPECTED_EOF;
    }
    if ((j = strlen(readbuf)) < 81) {  // Pad end with blanks for copying.
      while (j < 81) 
        readbuf[j] = ' ';
    }
    for (iword = 0; iword < 10 && i < jh->ncon; iword++, i++) {
      strncpy(jh->cnam[i], &readbuf[2 + iword*8], 6);
      jh->cnam[i][6] = '\0';
    }
  }

  // GROUP 1041: Constant values.
  if ((i =  ae_jpl_ascii_next_group(group, "GROUP   1041", jh->fp))) {
    fclose(jh->fp);
    return i;
  }
  if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp) == NULL) {
    fclose(jh->fp);
    return -AE_RET_UNEXPECTED_EOF;
  }
  jh->nval = atoi(readbuf);
  if (jh->nval != jh->ncon) {
    fclose(jh->fp);
    return -AE_RET_BAD_JPL_HEADER_NUM;
  }

  // Now read constant values, 3 per line, 26 characters each.
  for (i = 0; i < jh->ncon; i += 3) {
    if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp) == NULL) {
      fclose(jh->fp);
      return -AE_RET_UNEXPECTED_EOF;
    }

    for (j = 0; j < strlen(readbuf); j++) {
      if (tolower(readbuf[j]) == 'd') 
        readbuf[j] = 'E';   // Exponent is 'E'.
    }
    sscanf(readbuf, "%lE %lE %lE", &jh->cval[i], &jh->cval[i+1], 
           &jh->cval[i+2]);
  }
  
  // GROUP 1050: Constant values.
  if ((i = ae_jpl_ascii_next_group(group, "GROUP   1050", jh->fp))) {
    fclose(jh->fp);
    return i;
  }
  for (i =0; i < 3; i++) {
    // Read line of 13 6-digit integers.
    if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp) == NULL) {
      fclose(jh->fp);
      return -AE_RET_UNEXPECTED_EOF;
    }
    for (j = 0; j < 12; j++)
      jh->ipt[j][i] = atoi(&readbuf[6 * j]);
    jh->lpt[i] = atoi(&readbuf[6 * 12]);
  }

  // If there are no coefficients for an ipt[i][] object (i.e., ipt[i][1]==0),
  // then ipt[i][0] should contain the value of the next available coefficient
  // number rather than 0, as per communication of Myles Standish to Paul Hardy
  // on preferred format of ephemeris headers.
  //
  // If there are no libration coefficients (i.e., lpt[1]==0), then lpt[0]
  // should contain the value of the next available coefficient number rather
  // than 0 as well, as per the same communication from Myles Standish.
  //
  // First set j to maximum index into ipt[] that has coefficients.
  for (i = 1, j = 0; i < 12; i++)
    if (jh->ipt[i][1] > 0 && jh->ipt[i][0] > j)
      j = i;
  // Now set j to next available index count.
  if (jh->lpt[1] > 0 && jh->lpt[0] > j)
    j = jh->lpt[1] + jh->lpt[1] * jh->lpt[2] * 3;
  else
    j = jh->ipt[j][0] + jh->ipt[j][1] * jh->ipt[j][2] * 
        (j == 11 ? 2 : 3);
  for (i = 1; i < 12; i++)
    if (jh->ipt[i][0] == 0) jh->ipt[i][0] = j;
  if (jh->lpt[0] == 0)
    jh->lpt[0] = j;

  // Set the maximum number of Chebyshev coefficients possible for this file,
  // to initialize position and velocity Chebyshev coefficient arrays during
  // Chebyshev interpolation.
  jh->maxcheby = 0;
  for (i = 0; i < 12; i++)
    if (jh->ipt[i][1] > jh->maxcheby)
      jh->maxcheby = jh->ipt[i][1];
  if (jh->lpt[1] > jh->maxcheby)
    jh->maxcheby = jh->lpt[1];

  jh->au = 0.0;
  jh->emrat = 0.0;
  jh->numde = 0;
  for (i = 0; i < jh->ncon; i++) {
    if (strncmp(jh->cnam[i], "AU    ", 6) == 0)
      jh->au = jh->cval[i];
    else if (strncmp(jh->cnam[i], "EMRAT ", 6) == 0)
      jh->emrat = jh->cval[i];
    else if (strncmp(jh->cnam[i], "DENUM ", 6) == 0)
      jh->numde = jh->cval[i];
    else if (strncmp(jh->cnam[i], "CLIGHT", 6) == 0)
      jh->clight = jh->cval[i];
    else if (strncmp(jh->cnam[i], "LENUM ", 6) == 0)
      jh->numle = jh->cval[i];
  }
  if (jh->numle == 0)
    jh->numle = jh->numde;

  // GROUP 1070: Constant values.
  if ((i = ae_jpl_ascii_next_group(group, "GROUP   1070", jh->fp))) {
    fclose(jh->fp);
    return -AE_RET_UNEXPECTED_EOF;
  }

  fclose(jh->fp);

  return 0;
}


//------------------------------------------------------------------------------
//! Read a block of data coefficients from a JPL ASCII ephemeris file.
//! 
//! \param jh The handle for this file.  (The function ae_jpl_init() should be 
//!           called first.)
//!
//! \return The number of coefficients read, or 0 on EOF.  If there is an error,
//!         a negative error code from #ae_retcode_t is returned.
//------------------------------------------------------------------------------

int ae_jpl_read_ascii_block(struct ae_jpl_handle_t *jh) {
  char readbuf[AE_JPL_MAXLINE + 1], *dummy;
  int i, j, datalines, datapoints;
  double val1, val2, val3;

  // First line in an ASCII block will be the block number, followed by the 
  // number of coefficients.
  datalines = 0;  // Not reported, but leave in for debugging.
  datapoints = 0;
  if (fgets(readbuf, AE_JPL_MAXLINE, jh->fp) != NULL) {
    sscanf(readbuf, "%d %d", &i, &j);
    if (j != jh->ncoeff)
      return -AE_RET_BAD_JPL_BLOCK;
   
    datalines++;
    for (i = 0; i < jh->ncoeff && !feof(jh->fp); i += 3) {
      dummy = fgets(readbuf, AE_JPL_MAXLINE, jh->fp);
      
      for (j = 0; j < strlen(readbuf); j++)
        if (tolower(readbuf[j]) == 'd')  readbuf[j] = 'e';
      datalines++;
      
      // This is horrible, but use "%le" here and "%lE in the other ASCII data 
      // routine (ephcom_readascii_header) so gcc won't try to store the formats
      // in the same location and write to them.  (Problem with gcc not acting 
      // like K&R without -traditional flag and without -fwritable-strings 
      // flag.)
      sscanf(readbuf, " %le %le %le", &val1, &val2, &val3);
      jh->datablock[i] = val1;
      datapoints++;
      if ((i + 1) < jh->ncoeff) {
        jh->datablock[i + 1] = val2;
        datapoints++;
        if ((i + 2) < jh->ncoeff) {
          jh->datablock[i + 2] = val3;
          datapoints++;
        }
      }
    }
  }

  return datapoints;
}


//------------------------------------------------------------------------------
//! Read a JPL Ephemeris header in binary format.
//!
//! \param path The location of the ephemeris file.
//! \param jh For returning the header information.
//!
//! \return 0 on success; otherwise a negative error code from #ae_retcode_t.
//------------------------------------------------------------------------------

int ae_jpl_read_bin_header(const char *path, struct ae_jpl_handle_t *jh) {
  int i, j, k;

  if ((jh->fp = fopen(path, "r")) == NULL)
    return -AE_RET_BAD_PATH;

  // Read the title lines.
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 84; j++) {
      if ((jh->ttl[i][j] = fgetc(jh->fp)) == EOF) {
        fclose(jh->fp);
        return -AE_RET_UNEXPECTED_EOF;
      }
    }
    if (i == 0 && strncmp(jh->ttl[0], "JPL ", 4) &&
                  strncmp(jh->ttl[0], "DAF/SPK", 7) &&
                  strncmp(jh->ttl[0], "NAIF/DAF", 8)) {
      fclose(jh->fp);
      return -AE_RET_BAD_JPL_NOT_BIN;
    }
    jh->ttl[i][j] = '\0';
  }

  // Read constant names.
  for (i = 0; i < 400; i++) {
    for (j = 0; j < 6; j++) {
      if ((jh->cnam[i][j] = fgetc(jh->fp)) == EOF) {
        fclose(jh->fp);
        return -AE_RET_UNEXPECTED_EOF;
      }
    }
    jh->cnam[i][j] = '\0';
  }

  // Read ephemeris start epoch, stop epoch, and step size (in Julian Days).
  for (i = 0; i < 3; i++) {
    if ((j = ae_jpl_get_double(jh->fp, &(jh->ss[i])))) {
      fclose(jh->fp);
      return j;
    }
  }
  
  // Read NCON, AU, EMRAT.
  if ((i = ae_jpl_get_int(jh->fp, &(jh->ncon)))) {
    fclose(jh->fp);
    return i;
  }
  if ((i = ae_jpl_get_double(jh->fp, &(jh->au)))) {
    fclose(jh->fp);
    return i;
  }
  if ((i = ae_jpl_get_double(jh->fp, &(jh->emrat)))) {
    fclose(jh->fp);
    return i;
  }
  jh->nval  = jh->ncon;

  // Read indexes for coefficients in data block.  Written in transposed order 
  // (Fortran and C matrices are transposed).
  for (i = 0; i < 12; i++) {
    for (j = 0; j < 3; j++) {
      if ((k = ae_jpl_get_int(jh->fp, &(jh->ipt[i][j])))) {
        fclose(jh->fp);
        return k;
      }
    }
  }

  // Get ephemeris number.
  if ((i = ae_jpl_get_int(jh->fp, &(jh->numde)))) {
    fclose(jh->fp);
    return i;
  }
  for (i = 0; i < 3; i++) {
    if ((j = ae_jpl_get_int(jh->fp, &(jh->lpt[i])))) {
      fclose(jh->fp);
      return j;
    }
  }

  // If there are no coefficients for an ipt[i][] object (i.e., ipt[i][1]==0),
  // then ipt[i][0] should contain the value of the next available coefficient
  // number rather than 0, as per communication of Myles Standish to Paul Hardy
  // on preferred format of ephemeris headers.
  //
  // If there are no libration coefficients (i.e., lpt[1]==0), then lpt[0]
  // should contain the value of the next available coefficient number rather
  // than 0 as well, as per the same communication from Myles Standish.
  //
  // First set j to maximum index into ipt[] that has coefficients.
  for (i = 1, j = 0; i < 12; i++) {
    if (jh->ipt[i][1] > 0 && jh->ipt[i][0] > j)
      j = i;
  }

  // Now set j to next available index count.
  if (jh->lpt[1] > 0 && jh->lpt[0] > j)
    j = jh->lpt[1] + jh->lpt[1] * jh->lpt[2] * 3;
  else
    j = jh->ipt[j][0] + jh->ipt[j][1] * jh->ipt[j][2] * 
        (j == 11 ? 2 : 3);
  for (i = 1; i < 12; i++) {
    if (jh->ipt[i][0] == 0) 
      jh->ipt[i][0] = j;
  }
  if (jh->lpt[0] == 0) 
    jh->lpt[0] = j;

  // Set the maximum number of Chebyshev coefficients possible for this file,
  // to initialize position and velocity Chebyshev coefficient arrays during
  // Chebyshev interpolation.
  jh->maxcheby = 0;
  for (i = 0; i < 12; i++) {
    if (jh->ipt[i][1] > jh->maxcheby)
      jh->maxcheby = jh->ipt[i][1];
  }
  if (jh->lpt[1] > jh->maxcheby)
    jh->maxcheby = jh->lpt[1];

  // Calculate number of coefficients, starting with highest index into a data 
  // block (stored in j).
  for (i = 1, j = 0; i < 12; i++) {
    if (jh->ipt[i][1] > 0 && jh->ipt[i][0] > jh->ipt[j][0]) 
      j = i;
  }

  // Now see if the starting point we found is lower than where lpt[] starts.  
  // If not, use lpt[] for largest value.
  if (jh->lpt[1] > 0 && jh->lpt[0] > jh->ipt[j][0]) {
    jh->ncoeff = jh->lpt[0] - 1 +   // Starting point.
                 (jh->lpt[1] *      // Coefficients per coordinate.
                 jh->lpt[2]) *      // Subblocks per block.
                 3;                 // Coordinates.
  }
  else {
    jh->ncoeff = jh->ipt[j][0] - 1 +  // Starting point.
                 (jh->ipt[j][1] *     // Coefficients per coordinate.
                 jh->ipt[j][2]) *     // Subblocks per block.
                 (j == 11 ? 2 : 3);   // Coordinates.
   }

  jh->ksize = jh->ncoeff + jh->ncoeff; // KSIZE = 2 * NCOEFF always.

  // Skip to second block in file.
  fseek(jh->fp, jh->ncoeff * 8, SEEK_SET);
   
  // Read ephemeris constants.
  for (i = 0; i < jh->ncon; i++) {
    if ((j = jh->cval[i] = ae_jpl_get_double(jh->fp, &(jh->cval[i])))) {
      fclose(jh->fp);
      return j;
    }
    if (strncmp(jh->cnam[i], "NCOEFF", 6) == 0) {
      jh->ncoeff = jh->cval[i];
      jh->ksize  = 2 * jh->cval[i];
    }
    else if (strncmp(jh->cnam[i], "LENUM ", 6) == 0)
      jh->numle = jh->cval[i];
  }
  if (jh->numle == 0)
    jh->numle = jh->numde;

  return 0;
}


//------------------------------------------------------------------------------
//! Read a JPL Ephemeris data block in binary format.
//! The block number ranges from 0 on up (starting at first data block, after 
//! the 2 header blocks).  
//!
//! \param jh The handle for this file.  (The function ae_jpl_init() should be 
//!           called first.)
//! \param blocknum The data block number, starting at 0.
//!
//! \return The number of coefficients read, or 0 on EOF.  If there is an error,
//!         a negative error code from #ae_retcode_t is returned.
//------------------------------------------------------------------------------

int ae_jpl_read_bin_block(struct ae_jpl_handle_t *jh, int blocknum) { 
  int i;
  long filebyte;

  filebyte = (blocknum + 2) * jh->ncoeff * 8;   // 8 bytes per coefficient.
  fseek(jh->fp, filebyte, SEEK_SET);
  for (i = 0; !feof(jh->fp) && i < jh->ncoeff; i++)
    ae_jpl_get_double(jh->fp, &(jh->datablock[i]));
  if (i < jh->ncoeff && feof(jh->fp)) 
    i = 0;

  return i;
}


//------------------------------------------------------------------------------
//! Get the next group header in an ASCII file.
//! Group header lines are 12 characters long, of the form "GROUP   nnnn".
//!
//! \param group The GROUP header we read.
//! \param expected The header we expected.
//! \param fp The file pointer to read.
//!
//! \return 0 on success; otherwise a negative error code from #ae_retcode_t.
//------------------------------------------------------------------------------

int ae_jpl_ascii_next_group(char *group, char *expected, FILE *fp) {
  char readbuf[AE_JPL_MAXLINE + 1];

  if (fgets(readbuf, AE_JPL_MAXLINE, fp) == NULL)
    return -AE_RET_UNEXPECTED_EOF;
  if (fgets(readbuf, AE_JPL_MAXLINE, fp) == NULL)  // "GROUP   dddd\n".
    return -AE_RET_UNEXPECTED_EOF;
  strncpy(group, readbuf, 12);
  group[12] = '\0';

  if (strncmp(group, expected, 12) != 0)
    return -AE_RET_BAD_JPL_HEADER_GROUP;

  if (fgets(readbuf, AE_JPL_MAXLINE, fp) == NULL)  // Blank Line.
    return -AE_RET_UNEXPECTED_EOF;

  return 0;
}


//------------------------------------------------------------------------------
//! Read in a double precision value from a JPL binary file.
//! Bytes will be swapped if necessary to match host order (Little- or DEC- 
//! Endian).  On Intel 80x86 the bytes will get swapped, on Motorola or SPARC
//! they won't.
//!
//! \param fp The file from which to read the value.
//! \param val For returning the value.
//!
//! \return 0 on success; -AE_RET_UNEXPECTED_EOF if EOF is encountered.
//------------------------------------------------------------------------------

int ae_jpl_get_double(FILE *fp, double *val) {
  unsigned char ch[8];
   
  if (!fread(ch, 1, 8, fp))
    return -AE_RET_UNEXPECTED_EOF;
  ae_gnulliver64c(ch);
  memcpy((void *)val, (const void *)ch, (size_t)8);

  return 0;
}


//------------------------------------------------------------------------------
//! Read in an integer (4-byte) value from a JPL binary file.
//! Bytes will be swapped if necessary to match host order (Little- or DEC- 
//! Endian).  On Intel 80x86 the bytes will get swapped, on Motorola or SPARC
//! they won't.
//!
//! \param fp The file from which to read the value.
//! \param val For returning the value.
//!
//! \return 0 on success; -AE_RET_UNEXPECTED_EOF if EOF is encountered.
//------------------------------------------------------------------------------

int ae_jpl_get_int(FILE *fp, int *val) {
  unsigned u;

  if (!fread(&u, 4, 1, fp))
    return -AE_RET_UNEXPECTED_EOF;
  *val = (int)ae_gnulliver32(u);

  return 0;
}


//------------------------------------------------------------------------------
//! Get object positions from a JPL ephemeris file.
//!
//! \param jh A handle initialised by ae_jpl_init().
//! \param jd_tt The Julian date in TT.
//! \param obj_num The object for which to get coordinates.  If a planetary 
//!                ephemeris file is being used, one can use the predefined
//!                #ae_ss_bodies_t constants for clarity.
//! \param r For returning the rectangular position coordinates, in AU.
//! \param v For returning the rectangular velocity, in AU per day.  Set to
//!          NULL if you do not require this information.
//! \param is_planetary If 0, assume these are not planetary ephemerides and
//!                     barycentric coordinates are returned.  Otherwise, it 
//!                     will be assumed that a Solar System ephemeris file is 
//!                     being used and that the object numbers are as in 
//!                     #ae_ss_bodies_t; heliocentric coordinates will be
//!                     returned.
//!
//! \return 0 on success; on failure, a negative code from #ae_retcode_t.
//------------------------------------------------------------------------------

int ae_jpl_get_coords(struct ae_jpl_handle_t *jh, double jd_tt, int obj_num, 
                      double r[], double v[], char is_planetary) {
  int i, blocknum;
  double filetime;
  double obj[6], sun[6], embary[6], moon_emb[6];

  if (jd_tt < jh->ss[0] || jd_tt > jh->ss[1])
    return -AE_RET_BAD_JPL_DATE;
   
  filetime = jd_tt - jh->ss[0];    // Days from start of file.
  blocknum = (int)(filetime / jh->ss[2]); // Data block in file.

  // Read the data block that contains coefficients for desired date.
  if (jh->is_bin) {
    if (!ae_jpl_read_bin_block(jh, blocknum))
      return -AE_RET_UNEXPECTED_EOF;
  }
  else {
    fprintf(stderr, "ASCII reading not yet supported.\n");
    exit(0);
  }

  if (is_planetary) {
    // Get the sun so that we can return things in heliocentric coordinates.
    if ((i = ae_jpl_cheby(jh, jd_tt, AE_SS_SUN, sun, 1)))
      return i;

    // Find the barycentric coordinates.  There are a couple of special cases.
    switch (obj_num) {
      case AE_SS_EARTH: case AE_SS_MOON:
        // Get the earth-moon barycentre and the position of the moon relative
        // to this barycentre.
        if ((i = ae_jpl_cheby(jh, jd_tt, AE_SS_EMBARY, embary, 1)))
          return i;
        if ((i = ae_jpl_cheby(jh, jd_tt, AE_SS_MOON_EMB, moon_emb, 1)))
          return i;
        for (i = 0; i < 6; i++) {
          switch (obj_num) {
            case AE_SS_EARTH:
              obj[i] = embary[i] - moon_emb[i] / (1.0 + AE_E_M_RAT);
              break;
            case AE_SS_MOON:
              obj[i] = moon_emb[i] * (AE_E_M_RAT) / (1.0 + AE_E_M_RAT) +
                       embary[i];
              break;
            default:
              fprintf(stderr, "Should never have reached here.  Fix the "
                              "code!\n");
              exit(0);
          }
        }
        break;

      case AE_SS_SSBARY:
        // In barycentric coordinates, the barycentre is at the origin.
        for (i = 0; i < 6; i++)
          obj[i] = 0;
        break;

      default:
        if (obj_num < 0 || obj_num >= AE_N_SS_BODIES_JPL)
          return -AE_RET_BAD_OBJ_INDEX;
        if ((i = ae_jpl_cheby(jh, jd_tt, obj_num, obj, 1)))
          return i;
        break;
    }

    // Convert to heliocentric coordinates.
    for (i = 0; i < 6; i++)
      obj[i] -= sun[i];
  }
  else {
    if ((i = ae_jpl_cheby(jh, jd_tt, obj_num, obj, 0)))
      return i;
  }

  // Seperate positions und velocities and convert to AU.
  for (i = 0; i < 3; i++)
    r[i] = obj[i] / AE_AU;
  if (v != NULL) {
    for (i = 0; i < 3; i++)
      v[i] = obj[i + 3] / AE_AU;
  }

  return 0;
}


//------------------------------------------------------------------------------
//! Interpolate at a point using Chebyshev coefficients.
//!
//! \param jh A handle initialised with ae_jpl_init().
//! \param jd_tt The Julian date in TT.
//! \param obj_num The object index.
//! \param *pv Array to hold position in 1st half, velocity in 2nd.
//! \param is_planetary If non-zero, then it is assumed these are planetary
//!                     ephemerides.
//!
//! \return 0 on success; on failure, a negative error code from #ae_retcode_t.
//------------------------------------------------------------------------------

int ae_jpl_cheby(struct ae_jpl_handle_t *jh, double jd_tt, int obj_num, 
                 double *pv, char is_planetary) {
  int i, j, subinterval, dataoffset, ncoords, ncoeffs;
  double blocktime, span, subtime, x, *y;
  static double twox, lastx = 2.0;

  blocktime = jd_tt - jh->datablock[0];
  if (is_planetary && obj_num == AE_SS_NUTATION)
    ncoords = 2;
  else
    ncoords = 3;
  ncoeffs = jh->ipt[obj_num][1];
  span = jh->ss[2] / jh->ipt[obj_num][2];    // Days / subinterval.
  subinterval = (int)((jd_tt - jh->datablock[0]) / span);
  dataoffset = jh->ipt[obj_num][0] - 1 + ncoords * jh->ipt[obj_num][1] * 
               subinterval;
  subtime = blocktime - subinterval * jh->ss[2] / jh->ipt[obj_num][2];
    
  // Divide days in this subblock by total days in subblock to get interval 
  // [0,1].  The right part of the expression will evaluate to a whole number:
  // subinterval lengths are all integer multiples of days in a block (all 
  // powers of 2).
  x = subtime / span;
  x = (x + x) - 1.0;
  if (x < -1.0 || x > 1.0)
    return -AE_RET_BAD_JPL_CHEBY_TIME;
  y = &(jh->datablock[dataoffset]);
  
  if (lastx != x) {
    lastx = x;
    twox = x + x;   // For Chebyshev recursion.
    
    // Initialize position polynomial coefficients.
    jh->pc[0] = 1.0;    // Chebyshev T[0](x) = 1.
    jh->pc[1] = x;      // Chebyshev T[1](x) = x.
    for (i = 2; i < jh->maxcheby; i++) {
      jh->pc[i] = twox * jh->pc[i-1] - jh->pc[i-2];
      
      // Resolve bug with gcc generating -0.0 (also makes the smallest 
      // represented number equal to zero).
      if (jh->pc[i] * jh->pc[i] == 0.0)
        jh->pc[i] = 0.0;
    }
   
    // Initialize derivative polynomial coefficients
    jh->vc[0] = 0.0;          // d(1)/dx = 0 
    jh->vc[1] = 1.0;          // d(x)/dx = 1
    jh->vc[2] = twox + twox;  // d(2x^2 - 1)/dx = 4x
    for (i = 3; i < jh->maxcheby; i++)
      jh->vc[i] = twox * jh->vc[i - 1] + jh->pc[i - 1] + jh->pc[i - 1] - 
                         jh->vc[i - 2];
  }

  // Interpolate to get position for each component.
  for (i = 0; i < ncoords; i++) { 
    pv[i] = 0.0;
    for (j = ncoeffs - 1; j >= 0; j--)
      pv[i] += jh->pc[j] * y[i * ncoeffs + j];
  }

  // Interpolate velocity.
  for (i = 0; i < ncoords; i++) {
    pv[ncoords + i] = 0.0;
    for (j = ncoeffs - 1; j >= 0; j--)
      pv[ncoords + i] += jh->vc[j] * y[i * ncoeffs + j];
    pv[ncoords + i] *= 2.0 / span;
  }

  return 0;
}
