//! \file parse_iers.c
//! Parse IERS files to create object file for AEPHEM.
//! This file becomes a stand-alone executable.  It reads in the "finals.all",
//! "deltat.data" and "historic_deltat.data" files from 
//! http://maia.usno.navy.mil/ and compiles them into an output file, which can
//! be used to replace time_data.c, which AEPHEM uses for its DUT1 and delta T
//! calculations.  One can use this to keep time_data.c up-to-date as these
//! files are updated on the IERS server.
//==============================================================================
// AEPHEM - an astronomical ephemeris and reduction library.
// Copyright 2008 Adam Hincks.
//
// This file is part of AEPHEM.
//
// AEPHEM is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AEPHEM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AEPHEM.  If not, see <http://www.gnu.org/licenses/>.
//==============================================================================

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void usage();

int main(int argc, char *argv[]) {
  int i, n, year, last_year, last_last_year, first_mjd, last_mjd, mjd;
  char tmpstr[512];
  float dt, yearf;
  double dut1, c1, c2, c3;
  FILE *fp;

  if (argc != 7)
    usage();

  // Preliminary output.
  printf("//! \\file time_data.c\n"
    "//! Look-up tables for ae_delta_t() and ae_dut1().\n"
    "//! This file was generated by parse_iers on " __DATE__ " at " __TIME__
      ".\n"
    "//========================================================================"
      "======\n"
    "// AEPHEM - an astronomical ephemeris and reduction library.\n"
    "// Copyright 2008 Adam Hincks.\n"
    "//\n"
    "// This file is part of AEPHEM.\n"
    "//\n"
    "// AEPHEM is free software: you can redistribute it and/or modify\n"
    "// it under the terms of the GNU General Public License as published by\n"
    "// the Free Software Foundation, either version 3 of the License, or\n"
    "// (at your option) any later version.\n"
    "//\n"
    "// AEPHEM is distributed in the hope that it will be useful,\n"
    "// but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "// GNU General Public License for more details.\n"
    "//\n"
    "// You should have received a copy of the GNU General Public License\n"
    "// along with AEPHEM.  If not, see <http://www.gnu.org/licenses/>.\n"
    "//========================================================================"
    "======\n"
    "\n"
    "//! A table of delta_T = Ephemeris Time - Universal Time over many "
      "years.\n"
    "//! There is one value for each year, starting at #ae_delta_t_start_year"
      "and\n"
    "//! ending at #ae_delta_t_end_year.\n"
    "//!\n"
    "//! The tabulated range is 1620.0 onwards.  The values from 1620 to "
      "1973\n"
    "//! were taken from The Astronomical Almanac, page K8.  The values from "
      "2004\n"
    "//! onwards (including future projections) are taken from\n"
    "//! http://maia.usno.navy.mil.  Bessel's interpolation formula is "
      "implemented to\n"
    "//! obtain fourth order interpolated values at intermediate times.\n"
    "//!\n"
    "//! Updated deltaT predictions can be obtained from this network "
      "archive:\n"
    "//!    - http://maia.usno.navy.mil\n"
    "//!\n"
    "//! N.B.: Stephenson and Morrison's table starts at the year 1630.  The \n"
    "//! Chapronts' table does not agree with the Almanac prior to 1630.  The "
      "actual \n"
    "//! accuracy decreases rapidly prior to 1780.\n"
    "short ae_delta_t_tab[] = {\n"
    "  // ---------- 1620 - 1972 ----------\n"
    "  12400, 11900, 11500, 11000, 10600, 10200, 9800, 9500, 9100, 8800,\n"
    "  8500, 8200, 7900, 7700, 7400, 7200, 7000, 6700, 6500, 6300,\n"
    "  6200, 6000, 5800, 5700, 5500, 5400, 5300, 5100, 5000, 4900,\n"
    "  4800, 4700, 4600, 4500, 4400, 4300, 4200, 4100, 4000, 3800,\n"
    "  3700, 3600, 3500, 3400, 3300, 3200, 3100, 3000, 2800, 2700,\n"
    "  2600, 2500, 2400, 2300, 2200, 2100, 2000, 1900, 1800, 1700,\n"
    "  1600, 1500, 1400, 1400, 1300, 1200, 1200, 1100, 1100, 1000,\n"
    "  1000, 1000, 900, 900, 900, 900, 900, 900, 900, 900,\n"
    "  900, 900, 900, 900, 900, 900, 900, 900, 1000, 1000,\n"
    "  1000, 1000, 1000, 1000, 1000, 1000, 1000, 1100, 1100, 1100,\n"
    "  1100, 1100, 1100, 1100, 1100, 1100, 1100, 1100, 1100, 1100,\n"
    "  1100, 1100, 1100, 1100, 1200, 1200, 1200, 1200, 1200, 1200,\n"
    "  1200, 1200, 1200, 1200, 1300, 1300, 1300, 1300, 1300, 1300,\n"
    "  1300, 1400, 1400, 1400, 1400, 1400, 1400, 1400, 1500, 1500,\n"
    "  1500, 1500, 1500, 1500, 1500, 1600, 1600, 1600, 1600, 1600,\n"
    "  1600, 1600, 1600, 1600, 1600, 1700, 1700, 1700, 1700, 1700,\n"
    "  1700, 1700, 1700, 1700, 1700, 1700, 1700, 1700, 1700, 1700,\n"
    "  1700, 1700, 1600, 1600, 1600, 1600, 1500, 1500, 1400, 1400,\n"
    "  1370, 1340, 1310, 1290, 1270, 1260, 1250, 1250, 1250, 1250,\n"
    "  1250, 1250, 1250, 1250, 1250, 1250, 1250, 1240, 1230, 1220,\n"
    "  1200, 1170, 1140, 1110, 1060, 1020, 960, 910, 860, 800,\n"
    "  750, 700, 660, 630, 600, 580, 570, 560, 560, 560,\n"
    "  570, 580, 590, 610, 620, 630, 650, 660, 680, 690,\n"
    "  710, 720, 730, 740, 750, 760, 770, 770, 780, 780,\n"
    "  788, 782, 754, 697, 640, 602, 541, 410, 292, 182,\n"
    "  161, 10, -102, -128, -269, -324, -364, -454, -471, -511,\n"
    "  -540, -542, -520, -546, -546, -579, -563, -564, -580, -566,\n"
    "  -587, -601, -619, -664, -644, -647, -609, -576, -466, -374,\n"
    "  -272, -154, -2, 124, 264, 386, 537, 614, 775, 913,\n"
    "  1046, 1153, 1336, 1465, 1601, 1720, 1824, 1906, 2025, 2095,\n"
    "  2116, 2225, 2241, 2303, 2349, 2362, 2386, 2449, 2434, 2408,\n"
    "  2402, 2400, 2387, 2395, 2386, 2393, 2373, 2392, 2396, 2402,\n"
    "  2433, 2483, 2530, 2570, 2624, 2677, 2728, 2778, 2825, 2871,\n"
    "  2915, 2957, 2997, 3036, 3072, 3107, 3135, 3168, 3218, 3268,\n"
    "  3315, 3359, 3400, 3447, 3503, 3573, 3654, 3743, 3829, 3920,\n"
    "  4018, 4117, 4223,\n");

  // Read in the delta T file.
  if ((fp = fopen(argv[2], "r")) == NULL) {
    printf("\n\nERROR:  could not open %s for reading.\n", argv[2]);
    exit(0);
  }
  
  // Count number of entries in the deltat.data file.
  for (n = 0, last_year = -1; fgets(tmpstr, 127, fp) != NULL; ) {
    sscanf(tmpstr, "%d", &year);
    if (year != last_year) {
      n++;
      last_year = year;
    }
  }
  if (--n < 0) {
    printf("\n\nERROR:  %s has no entries!\n", argv[2]);
    fclose(fp);
    exit(0);
  }
  rewind(fp);

  printf("  // --------- 1973 - %d ----------\n", 1973 + n);

  // Print the entries.
  printf("  ");
  for (n = 0, last_year = -1; fgets(tmpstr, 127, fp) != NULL;) {
    sscanf(tmpstr, "%d %*d %*d %f", &year, &dt);
    if (year != last_year) {
      if (n)
        printf(" ");
      printf("%d,", (int)(dt * 100.0));
      if (++n >= 10) {
        n = 0;
        printf("\n  ");
      }
      last_year = year;
    }
  }
  if (n)
    printf("\n");
  fclose(fp);

  // Read in the delta T predictions.
  if ((fp = fopen(argv[3], "r")) == NULL) {
    printf("\n\nERROR:  could not open %s for reading.\n", argv[2]);
    exit(0);
  }

  last_last_year = last_year;
  for (n = 0; fgets(tmpstr, 127, fp) != NULL;) {
    sscanf(tmpstr, "%f", &yearf);
    year = (int)yearf;
    if (!year)
      continue;
    if (year != last_year) {
      last_year = year;
      n++;
    }
  }
  if (n < 0) {
    printf("\n\nERROR:  %s has no entries!\n", argv[2]);
    fclose(fp);
    exit(0);
  }
  rewind(fp);

  printf("  // --------- %d - %d ----------\n", last_last_year + 1,
         last_last_year + n);
  
  // Print the entries.
  printf("  ");
  for (n = 0, last_year = last_last_year; fgets(tmpstr, 127, fp) != NULL;) {
    yearf = 0;
    sscanf(tmpstr, "%f %f", &yearf, &dt);
    year = (int)yearf;
    if (!year)
      continue;
    if (year != last_year) {
      if (n)
        printf(", ");
      printf("%d", (int)(dt * 100.0));
      if (++n >= 10) {
        n = 0;
        printf(",\n  ");
      }
      last_year = year;
    }
  }
  if (n)
    printf("\n");
  fclose(fp);

  printf("};\n"
    "\n"
    "//! The first year of delta T tabulation.\n"
    "const int ae_delta_t_start_year = 1620;\n"
    "//! The final year of delta T tabulation.\n"
    "const int ae_delta_t_end_year = %d;\n\n", year);

  printf("//! A table of DUT1, the difference between UT1 and UTC.\n"
    "//! The entries are in tenths of milliseconds, beginning at the Modified "
      "Julian\n"
    "//! Date given by #ae_dut1_first_mjd and incrementing every day "
      "thereafter.\n"
    "//! This format was chosen to minimise the amount of disc space used "
      "since\n"
    "//! short integers can be employed.\n"
    "//!\n"
    "//! The last year of the table consists of forecasted values.\n"
    "//!\n"
    "//! These data were taken from past IERS Bulletin A tables.  Values "
      "can\n"
    "//! be found online at http://maia.usno.navy.mil/search/search.html; the "
      "file\n"
    "//! finals.all compiles all values since 1973.\n"
    "short ae_dut1_tab[] = {\n");

  // Read from the dut1 file.
  if ((fp = fopen(argv[1], "r")) == NULL) {
    printf("\n\nERROR:  could not open %s for reading.\n", argv[2]);
    exit(0);
  }

  first_mjd = 0;
  for (n = 0, mjd = -1, dut1 = 0; fgets(tmpstr, 511, fp) != NULL; n++) {
    if (strlen(tmpstr) < 69) {
      printf("\n\nERROR: line in %s was too short.  Bailing.", argv[1]);
      fclose(fp);
      exit(0);
    }

    tmpstr[15] = '\0';
    last_mjd = atoi(&tmpstr[7]);

    if (mjd < 0)
      first_mjd = last_mjd;
    else {
      if (last_mjd <= mjd)
        continue;

      for (i = last_mjd + 1; i < mjd; i++)
        printf("%d, ", (int)(dut1 * 10000));
    }
    
    tmpstr[68] = '\0';
    if (!(dut1 = atof(&tmpstr[58]))) {
      last_mjd--;
      break;
    }
    if (n)
      printf(",");
    else
      printf(" ");
    if (n && !(n % 10)) {
      printf("\n  ");
    }
    else
      printf(" ");
    printf("%d", (int)(dut1 * 10000));

    mjd = last_mjd;
  }
  fclose(fp);

  if (!first_mjd) {
    printf("\n\nERROR: %s has no entries.\n", argv[1]);
    exit(0);
  }

  if (n % 10)
    printf("\n");


  printf("};\n"
    "\n"
    "//! The Modified Julian Date of the first entry in #ae_dut1_tab[].\n"
    "const double ae_dut1_first_mjd = %d;\n"
    "//! The Modified Julian Date of the last entry in #ae_dut1_tab[].\n"
    "const double ae_dut1_last_mjd = %d;\n", first_mjd, last_mjd);

  c1 = atof(argv[4]);
  c2 = atof(argv[5]);
  c3 = atof(argv[6]);

  printf("\n"
    "//! Offset for predicting DUT1 past the end of #ae_dut1_tab[].\n"
    "const double ae_dut1_predict_offset = %g;\n"
    "//! Scale for predicting DUT1 past the end of #ae_dut1_tab[].\n"
    "const double ae_dut1_predict_scale = %g;\n"
    "//! Reference MJD for predicting DUT1 past the end of #ae_dut1_tab[].\n"
    "const double ae_dut1_predict_mjd = %g;\n", c1, c2, c3);

  return 0;
}

void usage() {
  printf("Usage: parse_iers <finals.all> <deltat.data> <deltat.preds>\n");
  printf("                  <coeff1> <coeff2> <coeff3>\n");
  printf("Parse finals.all, deltat.data and deltat.preds, provided on\n"
         "http://maia.usno.navy.mil/.  The input coefficients should be taken "
           "from the\n"
         "most recent Bulletin A, also found on this website; they are from "
           "the line\n"
         "that gives a prediction for UT1-UTC.  For example, the line\n"
         "  UT1-UTC =  -.4948 -  .00082 (MJD - 54700) - (UT2-UT1)\n"
         "yields coeff1 = -0.4948, coeff2 = 0.00082 and coeff3 = 54700.\n"
         "The output is a C file suitable for replacing time_data.c.\n");
  exit(0);
}
